/* $Id: ncbithr.c,v 5.10 1996/09/04 19:37:32 epstein Exp $ */
/*****************************************************************************

    Name: ncbithr.c           

    Description: Internal functions for Multi-thread API library

    Author: Sergei Shavirin

   ***************************************************************************

                          PUBLIC DOMAIN NOTICE
              National Center for Biotechnology Information

    This software/database is a "United States Government Work" under the
    terms of the United States Copyright Act.  It was written as part of    
    the author's official duties as a United States Government employee
    and thus cannot be copyrighted.  This software/database is freely
    available to the public for use. The National Library of Medicine and
    the U.S. Government have not placed any restriction on its use or
    reproduction.

    Although all reasonable efforts have been taken to ensure the accuracy
    and reliability of the software and data, the NLM and the U.S.
    Government do not and cannot warrant the performance or results that
    may be obtained by using this software or data. The NLM and the U.S.
    Government disclaim all warranties, express or implied, including
    warranties of performance, merchantability or fitness for any
    particular purpose.

    Please cite the author in any work or product based on this material.

   ***************************************************************************

    Modification History:
           30 Apr 1996 - Shavirin Sergei  -  original written

    Bugs and restriction on use:

    Notes:
 * $Log: ncbithr.c,v $
 * Revision 5.10  1996/09/04  19:37:32  epstein
 * workaround for SGI4
 *
 * Revision 5.9  1996/09/04  14:34:44  kans
 * trivial fixes to allow compilation under Symantec C++
 *
 * Revision 5.8  1996/08/22  14:07:53  shavirin
 * Finaly NlmThreadsAvailable return TRUE for WIN32
 *
 * Revision 5.7  1996/08/07  17:07:58  shavirin
 * Added function NlmCPUNumber(void), that returns number of CPU
 * on calling host
 *
 * Revision 5.6  1996/08/02  19:52:38  vakatov
 * Most of the thread-handling code has been protected from being
 * compiled on the non-thread platforms
 *
 * Revision 5.5  1996/08/02  15:59:38  shavirin
 * Added support for POSIX threads. Implemented semaphores and
 * Read-Write lock objects for POSIX threads through POSIX mutexes
 *
 * Revision 5.4  1996/07/16  19:58:50  vakatov
 * [WIN32]  NlmThreadSelf() now returns handler rather than pseudo-handler.
 * [SOLARIS...] [POSIX...]  init_exit_arr() protected by "mutex" from.
 * [POSIX] "mutex"-related stuff implemented.
 * NlmThreadMemFree() auxiliary function added.
 * The "thread destructors" calling order has been changed from FIFO to LIFO.
 * Added NlmThreadCreateEx() function to specify the thread destructor
 * from outside the thread.
 * The program name property is to be automatially inherited by the threads.
 *
 * Revision 5.3  1996/07/09  18:13:37  vakatov
 * Added a basic MT-support for POSIX and WIN32-SDK(Windows-95 and Windows-NT)
 * API -- the thread creation, joining and terminating. The thread joining
 * (waiting for the terminating) interface and behaviour has been slightly
 * changed.
 * Tested the lately introduced "OnExit"-features for the new platforms.
 *
 * Revision 5.2  1996/07/05  21:09:22  vakatov
 * Added NlmThreadAddOnExit() and NlmThreadRemoveOnExit() functions allowing
 * the user to specify a set of functions to be automatically executed on the
 * thread exiting stage. In addition, the execution of NlmThreadExit() function
 * is guaranteed in the case when the thread's main function exits with
 * the "return" operator, without calling the NlmThreadExit() explicitly
 *
 * The concerning stuff is also implemented(however, not tested yet) for
 * POSIX and WIN32-SDK(MSVC++ and Borland C++) threads
 *
 * Tests are added at the end of the file(preprocessor variable TEST_MODULE)
 * to provide easy testing of the new(and some of the old) NCBI thread features
 * 

*****************************************************************************/

#include <ncbithri.h>
#include <ncbithr.h>

#if defined(SOLARIS_THREADS_AVAIL) || defined(POSIX_THREADS_AVAIL) || defined(WIN32)
#define NCBI_THREADS_AVAIL
#endif 


/***********************************************************************
 *  INTERNAL
 ***********************************************************************/


/* POSIX implementation suppose, that semaphore and RW objects
   must be implemented by user himself .... */

#if defined (POSIX_THREADS_AVAIL)

typedef struct psema_t {
	Int4       count;
	pthread_mutex_t    mutex;
	pthread_cond_t     cond;
} psema_t;

typedef struct rwlock_t {
	Int4       readers;
	pthread_mutex_t    mutex;
	pthread_cond_t     cond_r;
	pthread_cond_t     cond_w;
} rwlock_t;

#endif

/*
 *  Functions and data structures providing storage and execution of
 *  the "thread destructor functions"
 */

typedef struct
{
  TNlmThreadOnExit func;
  VoidPtr          arg;
}  TOnExitData;

typedef struct
{
  size_t      size;
  TOnExitData user[1];
}  TOnExitArray, PNTR TOnExitArrayPtr;
 

#if   defined (SOLARIS_THREADS_AVAIL)
static  thread_key_t thread_exit_key;
#elif defined (POSIX_THREADS_AVAIL)
static pthread_key_t thread_exit_key;
#elif defined (WIN32)
#ifdef _MSC_VER
static __declspec( thread ) TOnExitArrayPtr thread_exit_arr = NULL;
static __declspec( thread ) HANDLE          thread_self     = NULL;
#else
static TOnExitArrayPtr      __thread        thread_exit_arr = NULL;
static HANDLE               __thread        thread_self     = NULL;
#endif
#else
/* no threads -- no problem */
#endif


static Boolean init_exit_arr( void )
{
  static Boolean initialized = FALSE;

  if ( initialized )
    return TRUE;

#if   defined (SOLARIS_THREADS_AVAIL)
  {
    static mutex_t key_lock;
    mutex_lock( &key_lock );
    if (!initialized  &&
        thr_keycreate(&thread_exit_key, NULL) == 0)
      initialized = TRUE;
    mutex_unlock( &key_lock );
  }
#elif defined (POSIX_THREADS_AVAIL)
  {
    static pthread_mutex_t key_lock;
    pthread_mutex_lock( &key_lock );
    if (!initialized  &&
        pthread_key_create(&thread_exit_key, NULL) == 0)
      initialized = TRUE;
    pthread_mutex_unlock( &key_lock );
  }
#elif defined (WIN32)
  initialized = TRUE;
#endif

  return initialized;
}


static TOnExitArrayPtr get_exit_arr( void )
{
  if ( !init_exit_arr() )
    return NULL;

  {{
#if   defined (SOLARIS_THREADS_AVAIL)
  VoidPtr arr;
  if (thr_getspecific(thread_exit_key, &arr) != 0)
    return NULL;
  return (TOnExitArrayPtr)arr;
#elif defined (POSIX_THREADS_AVAIL)
  return (TOnExitArrayPtr)pthread_getspecific( thread_exit_key );
#elif defined (WIN32)
  return thread_exit_arr;
#else
  return NULL;
#endif
  }}
}


static void set_exit_arr(TOnExitArrayPtr arr_ptr)
{
#if   defined (SOLARIS_THREADS_AVAIL)
  if (    thr_setspecific(thread_exit_key, (VoidPtr)arr_ptr) != 0)
    abort();
#elif defined (POSIX_THREADS_AVAIL)
  if (pthread_setspecific(thread_exit_key, (VoidPtr)arr_ptr) != 0)
    abort();
#elif defined (WIN32)
  thread_exit_arr = arr_ptr;
#endif
}



/***********************************************************************
 *  EXTERNAL
 ***********************************************************************/


/* --------------------  NlmThreadAddOnExit  ----------------------
   Purpose:     Adds a function to be called on the thread termination
   Parameters:  The function "func" will be called with the "arg" argument
                when the thread is terminating;  then, the later registered
                "destroying function" will be called earlier (LIFO order)

   Returns:     TRUE on success...
   NOTE:        Must be called from inside the thread
  ------------------------------------------------------------------*/
extern Boolean NlmThreadAddOnExit(TNlmThreadOnExit func, VoidPtr arg)
{
#ifdef NCBI_THREADS_AVAIL
  TOnExitArrayPtr arr_ptr = get_exit_arr();
  size_t size = (arr_ptr == NULL) ? 0 : arr_ptr->size;

  size_t i = 0;
  while (i < size  &&  arr_ptr->user[i].func != NULL)
    i++;

  if (i == size)
    {
      TOnExitArrayPtr new_arr = (TOnExitArrayPtr)
        Nlm_MemNew(sizeof(TOnExitArray) + size * sizeof(TOnExitData));
      if (new_arr == NULL)
        return FALSE;

      new_arr->size = size + 1;
      if (size != 0)
        {
          Nlm_MemCpy(new_arr->user, arr_ptr->user,
                     size * sizeof(TOnExitData));
          Nlm_MemFree( arr_ptr );
        }
      arr_ptr = new_arr;
    }

  arr_ptr->user[i].func = func;
  arr_ptr->user[i].arg  = arg;
  set_exit_arr( arr_ptr );
  return TRUE;
#else
  return FALSE;
#endif
}


/* --------------------  NlmThreadRemoveOnExit  -------------------
   Purpose:     Removes the thread's termination function(s)
   Description: This function removes all previosly registered(for this
                thread) terminating functions whose attributes
                matches both "func" and "arg"
   NOTE:        Must be called from inside the thread
  ------------------------------------------------------------------*/
extern void NlmThreadRemoveOnExit(TNlmThreadOnExit func, VoidPtr arg)
{
#ifdef NCBI_THREADS_AVAIL
  TOnExitArrayPtr arr_ptr = get_exit_arr();
  size_t size = (arr_ptr == NULL) ? 0 : arr_ptr->size;

  size_t i;
  for (i = 0;  i < size;  i++)
    {
      if (arr_ptr->user[i].func == func  &&
          arr_ptr->user[i].arg  == arg)
        {
          Nlm_MemMove(arr_ptr->user + i, arr_ptr->user + i + 1,
                      (size - i - 1) * sizeof(TOnExitData));
          arr_ptr->user[size-1].func = NULL;
        }
    }
#endif
}


/* --------------------  NlmThreadWrapper  -------------------------
   Purpose:     To wrap the starting thread main function
   Parameters:  The user's main thread function with args and flags 
   Returns:     New thread handle. -- to be used to identify the thread.
  
   Description: This function creates new thread - independent process
                inside current process, that share most resources of the
                parent thread.
                
   NOTE:        This function must be used in NlmThreadCreate() only!
  ------------------------------------------------------------------*/

typedef struct
{
  TNlmThreadStart  func;
  VoidPtr          arg;
  TNlmThreadOnExit exit_func; 
  VoidPtr          exit_arg;
  CharPtr          progName;
} TNlmThreadWrapperArg, PNTR TNlmThreadWrapperArgPtr;


extern void NlmThreadMemFree(VoidPtr ptr)
{
  Nlm_MemFree( ptr );
}


static
#if defined(SOLARIS_THREADS_AVAIL) || defined(POSIX_THREADS_AVAIL)
VoidPtr
#elif defined(WIN32)
DWORD WINAPI
#endif
NlmThreadWrapper(VoidPtr wrapper_arg)
{
#ifdef NCBI_THREADS_AVAIL
  VoidPtr ret_value;
  TNlmThreadWrapperArgPtr user = (TNlmThreadWrapperArgPtr)wrapper_arg;
  TNlmThreadStart startFunc    = user->func;
  VoidPtr         startFuncArg = user->arg;

#ifdef WIN32
  /* duplicate and store an actual(not a pseudo-) handle to the thread */
  if ( !DuplicateHandle(GetCurrentProcess(), GetCurrentThread(),
                        GetCurrentProcess(), &thread_self,
                        0, FALSE, DUPLICATE_SAME_ACCESS) )
    abort();
#endif

  /* register the exiting function specified by the thread parent, if any */
  if (user->exit_func != NULL  &&
      !NlmThreadAddOnExit(user->exit_func, user->exit_arg))
    abort();  

  /* store the current program name */
  if (user->progName != NULL)
    {
      Nlm_SetAppProperty("ProgramName",    (void *)user->progName);
      NlmThreadAddOnExit(NlmThreadMemFree, (void *)user->progName);
    }

  Nlm_MemFree( wrapper_arg );

  /* run the user's thread function */
  ret_value = (*startFunc)( startFuncArg );

  /* the regular exit from the thread */
  NlmThreadExit( ret_value );

  /* should never reach here... */
  abort();
#endif

  return 0;
}

/* --------------------  NlmThreadCreate  ---------------------------
   Purpose:     To create new thread
  
   Parameters:  theStartFunction - address of the thread function
                *arg             - arguments for the new thread
                flags            - attributes for the thread: 
		                   THR_RUN (default) or THR_SUSPEND
   Returns:     New thread handle. Will be used to identify the thread. 
  
   Description: This function creates new thread - independent process
                inside current process, that share most resources of the
                parent thread.
                
   NOTE:         
  ------------------------------------------------------------------*/
TNlmThread NlmThreadCreate(
	       TNlmThreadStart theStartFunction,   /* address of the thread 
						      function */   
	       VoidPtr arg,         /* arguments for the new thread */
	       Int4    flags        /* attributes for the thread */
	   )
{
  return NlmThreadCreateEx(theStartFunction, arg, flags, NULL, NULL);
}


/* --------------------  NlmThreadCreateEx  -------------------------
   Purpose:     To create new thread and register its destructor
  
   Parameters:  "theStartFunction", "arg", "flags"  -- see NlmThreadCreate
                "exit_func", "exit_arg"  -- these two allow one to add a
                function to be called when the thread is exiting -- it's
                the only convenient way to do this from outside of thread
                (see also the NlmThreadAddOnExit() function to get know
                how to do it from the thread's inside).

   Returns:     New thread handle. Will be used to identify the thread. 
  
   Description: This function creates new thread - independent process
                inside current process, that share most resources of the
                parent thread.
                
   NOTE:        It's guaranteed that the "exit_func" (if non-NULL) will
                be called after any other exiting function specified
                from inside the thread
  ------------------------------------------------------------------*/
TNlmThread NlmThreadCreateEx(TNlmThreadStart  theStartFunction,
                             VoidPtr          arg,
                             Int4             flags,
                             TNlmThreadOnExit exit_func, 
                             VoidPtr          exit_arg            
                             )
{
  TNlmThread thread_handle = (TNlmThread)0;

#ifdef NCBI_THREADS_AVAIL
  /* initialize thread wrapper func */
  TNlmThreadWrapperArgPtr wrapper_data =
    Nlm_MemNew( sizeof(TNlmThreadWrapperArg) );
  if (wrapper_data == NULL)
    return NULL;
  wrapper_data->func      = theStartFunction;
  wrapper_data->arg       = arg;
  wrapper_data->progName  = (char *)GetAppProperty( "ProgramName" );
  wrapper_data->exit_func = exit_func;
  wrapper_data->exit_arg  = exit_arg;
  if (wrapper_data->progName != NULL)
    {
      char *progName = (char *)
        Nlm_MemNew(Nlm_StringLen(wrapper_data->progName) + 1);
      Nlm_StrCpy(progName, wrapper_data->progName);
      wrapper_data->progName = progName;
    }

  /* create new thread */
#if   defined(SOLARIS_THREADS_AVAIL)
  if(thr_create(NULL, 0, NlmThreadWrapper, (void *)wrapper_data, 
                (long)flags, (thread_t *)&thread_handle) != 0)
    thread_handle = (TNlmThread)0;

#elif defined(POSIX_THREADS_AVAIL)
  {
  pthread_attr_t attr;
  pthread_attr_init( &attr );
  if(pthread_create((pthread_t *)&thread_handle, &attr,
                    NlmThreadWrapper, (void *)wrapper_data) != 0)
    thread_handle = (TNlmThread)0;
  pthread_attr_destroy( &attr );
  }

#elif defined(WIN32)
  {
    DWORD thread_id;
    thread_handle = (TNlmThread)CreateThread(NULL, 0,
                                 NlmThreadWrapper, (LPVOID)wrapper_data,
                                 flags, &thread_id);
  }
#endif

  /* is OK? */
  if (thread_handle == (TNlmThread)0)
    {
      Nlm_MemFree( wrapper_data->progName );
      Nlm_MemFree( wrapper_data );
    }
#endif

  return thread_handle;
}


/* ----------------------  NlmThreadSelf  --------------------------
   Purpose:     To get thread handle 
  
   Parameters:  None
  
   Returns:     theThread - thread handle NULL if error
                an error
  
   Description: NlmThreadSelf() returns the handle of the calling thread
                
   NOTE:        
  -----------------------------------------------------------------*/

TNlmThread NlmThreadSelf(void)
{
#if   defined(SOLARIS_THREADS_AVAIL)
  return (TNlmThread)thr_self();
#elif defined(POSIX_THREADS_AVAIL)
  return (TNlmThread)pthread_self();
#elif defined(WIN32)
  return (TNlmThread)thread_self;
#else             /* default */
  return NULL;
#endif
}


/* ----------------------  NlmThreadJoin  --------------------------
   Purpose:     Wait for thread termination 
  
   Parameters:  INPUT:  "wait_for" -- thread handle to wait for termination
                OUTPUT: "status"   -- the thread's exit status
   Returns:     zero value on success;  non-zero value on error
  
   Description: Blocks the calling thread until the thread specified by
                wait_for terminates. The specified thread must be in the 
                current process. If wait_for is (TNlmThread)0, then
                NlmThreadJoin() waits for ANY thread in the process to
                terminate.
   NOTE:        Multiple threads cannot wait for the same thread to 
                terminate. One thread will return successfully and the 
                others will fail.
  -----------------------------------------------------------------*/
Int4 NlmThreadJoin(
		   TNlmThread  wait_for, /* handle of the thread to wait */
                   VoidPtr    *status    /* the thread's exit status     */
                  )
{
#if   defined(SOLARIS_THREADS_AVAIL)
  return thr_join((thread_t)wait_for, NULL, status);
#elif defined(POSIX_THREADS_AVAIL)
  return pthread_join((pthread_t)wait_for, status);
#elif defined(WIN32)
  HANDLE thread = (HANDLE)wait_for;
  WaitForSingleObject(thread, INFINITE);
  if (!GetExitCodeThread(thread, (LPDWORD)status)  ||
      (DWORD)*status == (DWORD)STILL_ACTIVE)
    {
      DWORD  err_code = GetLastError();
      if (err_code == 0)
        return -1;
      return (Int4)err_code;
    }

  CloseHandle( thread );
  return 0;
#else          /* default */
  return -1;
#endif
}

/* ----------------  NlmThreadSetConcurrency  ----------------------
   Purpose:     Set thread concurrency level
  
   Parameters:  new_level - concurrency level to set
  
   Returns:     0 is returned when successful. A non-zero value indicates 
                an error
  
   Description: This function permits the application to give the threads 
                system a hint, specified by new_level for the desired level 
		of concurrency. The actual number of simultaneously active
		threads may be larger or smaller than this number.
                
   NOTE: For Solaris Threads ONLY!
  -----------------------------------------------------------------*/
Int4 NlmThreadSetConcurrency(Int4 new_level)
{
#if defined(SOLARIS_THREADS_AVAIL)

  return(thr_setconcurrency(new_level));

#elif defined(WIN32)
  return 0;
#else
  return 0;
#endif
}

/* ----------------  NlmThreadGetConcurrency  ----------------------
   Purpose:     To get thread handle 
  
   Parameters:  None
  
   Returns:     Current value for the desired concurrency level
  
   Description: returns the current value for the desired concurrency
                level. The actual number of simultaneously active threads 
		may be larger or smaller than this number.
                
   NOTE: For Solaris Threads ONLY!
  -----------------------------------------------------------------*/
Int4 NlmThreadGetConcurrency(void)
{
#if defined(SOLARIS_THREADS_AVAIL)

  return(thr_getconcurrency());  

#elif defined(WIN32)
  return 0;
#else
  return 0;
#endif
}

/* ----------------------  NlmThreadContinue  -----------------------
   Purpose:     Resume execution of a suspended thread 
  
   Parameters:  theThread - thread handle to continue
  
   Returns:     0 is returned when successful. A non-zero value indicates 
                an error
  
   Description: Resume execution of the thread, that was initially
                started in suspended state
                
   NOTE:        Currently suspending supported only during initial
                creation of the thread
  -----------------------------------------------------------------*/
Int4 NlmThreadContinue(
		      TNlmThread theThread   /* thread handle to continue */
    )
{
#if defined(SOLARIS_THREADS_AVAIL)

  return(thr_continue((thread_t)theThread));  

#elif defined(WIN32)

  if(ResumeThread((HANDLE)theThread) != 0xFFFFFFFF)
	  return 0;
  else
	  return 0xFFFFFFFF;
#else
  return 0;
#endif
}


/* -------------------  NlmThreadExit  ------------------------------
   Purpose:      To terminate the calling thread    
  
   Parameters:   status - exit code for the calling thread
  
   Returns:      Nothing      
   
   Description:  This function terminates th calling thread. All thread-
                 specific data bindings are released. If the initial thread 
                 returns from main() then the process exits with a status
                 equal to the return value
                
   NOTE:
   -----------------------------------------------------------------*/
void NlmThreadExit(
		   VoidPtr status /* exit data of the calling thread */
                  )
{
  /* execute all terminating functions registered by user */
  TOnExitArrayPtr arr_ptr = get_exit_arr();
  long size = (arr_ptr == NULL) ? 0 : arr_ptr->size;
  while ( size-- )
    {
      if (arr_ptr->user[size].func != NULL)
        (*arr_ptr->user[size].func)( arr_ptr->user[size].arg );
    }

  if (arr_ptr != NULL)
    Free( arr_ptr );
  
  /* exiting from thread must clear "context" structure */
  Nlm_ReleaseAppContext();

#if defined(SOLARIS_THREADS_AVAIL)
  thr_exit     ( (void *) status );
#elif defined(POSIX_THREADS_AVAIL)
  pthread_exit ( (void *) status );
#elif defined(WIN32)
  CloseHandle( thread_self );
  ExitThread( (DWORD)  status );
#endif
}


/* -------------------  NlmThreadSetPriority  -----------------------
   Purpose:     To set priority of the specified thread   
  
   Parameters:  theThread - thread handle to set priority
                priority  - integer from 0 (the least priority)
                           to 127 (highest priority) 
  
   Returns:     Zero is returned when succesful. A non-zero value
                indicates error.
  
   Description: Each thread has a priority which it inherits from its
                creator. This function changes priority of "theThread" to
                the value specified by "priority".
                
   NOTE:  On Solaris POSIX SCHED_OTHER policy implemented - use this so far
   -----------------------------------------------------------------*/
Int4 NlmThreadSetPriority(
			 TNlmThread theThread, /* thread handle */
			 Int4 priority         /* priority level */
    )
{

#if defined(SOLARIS_THREADS_AVAIL)

  return (thr_setprio((thread_t)theThread, priority));  


#elif defined (POSIX_THREADS_AVAIL)

  struct sched_param param;	
  param.sched_priority = priority;
  return(pthread_setschedparam((pthread_t)theThread, SCHED_OTHER, &param));

#elif defined(WIN32)

  return (SetThreadPriority((HANDLE)theThread, priority));

#else
  return 0;
#endif
}

/* -------------------  NlmThreadGetPriority  -----------------------
   Purpose:      To get priority of the thread
  
   Parameters:   theThread - handle of the thread to get priority
                 priority  - storage for the priority of theThread
   Returns:      Zero is returned, when succesful. A non-zero value 
                 indicates error
  
   Description:  This function gets current value of priority of theThread
                
   NOTE: On Solaris POSIX only SCHED_OTHER policy implemented - use this so far
   -----------------------------------------------------------------*/
Int4 NlmThreadGetPriority(
                         TNlmThread theThread,  /* handle of the thread */
			 int *priority       /* storage for priority */
    )
{
#if defined(SOLARIS_THREADS_AVAIL)

  return (thr_getprio((thread_t)theThread, priority));  

  
#elif defined (POSIX_THREADS_AVAIL)

  struct sched_param param;	
  Int4 error;
  int policy; /* SCHED_OTHER the only policy for Solaris */

  if((error = pthread_getschedparam((pthread_t)theThread, &policy, &param)) == 0) { 
    *priority = param.sched_priority;
    return 0;
  } else {
    return error;
  }

#elif defined(WIN32)
  
  return (GetThreadPriority((HANDLE)theThread));

#else
  return 0;
#endif
}
/* ---------------------  NlmThreadKill  ----------------------------
   Purpose:     To terminate thread from another thread    
  
   Parameters:  theThread - handle of the thread to kill
  
   Returns:     Zero is returned when successful. A non-zero value
                indicates an error
  
   Description: This function terminates thread, specified as parameter.
                
   NOTE:        In Solaris we send signal "9" to the thread.
   -----------------------------------------------------------------*/
Int4 NlmThreadKill(
		  TNlmThread theThread   /* handle of the thread to kill */
    )
{
#if   defined(SOLARIS_THREADS_AVAIL)
  return thr_kill((thread_t)theThread, 9);  

#elif defined(POSIX_THREADS_AVAIL)
  return pthread_kill((pthread_t)theThread, 9);  

#elif defined(WIN32)
  return TerminateThread((HANDLE)theThread, 1);
#else
  return 0;
#endif
}

/********************************************************************/
/*                                                                  */
/* =========== Semaphore support for the thread API library ======= */
/*                                                                  */
/********************************************************************/


/* ---------------------  NlmSemaInit  ------------------------------
   Purpose:     To create new semaphore    
  
   Parameters:  count - number of threads, that may use semaphore
                in parallel
  
   Returns:     Semaphore handle
  
   Description: To create new semaphore with "count" simultaneous accesses.
                
   NOTE:        This semathore can be used to synchronize threads in this
                process only
   -----------------------------------------------------------------*/
TNlmSemaphore NlmSemaInit(
			  Uint4 count
              )
{
#if defined(SOLARIS_THREADS_AVAIL)

  sema_t *theSemaphore = (sema_t *)Malloc(sizeof(sema_t));

  if(theSemaphore == NULL || 
	              sema_init(theSemaphore, count, USYNC_THREAD, NULL) != 0)
    return NULL;
  else
   return (TNlmSemaphore)theSemaphore;

#elif defined(POSIX_THREADS_AVAIL)

  psema_t *theSemaphore; 
 	  
  if ((theSemaphore = (psema_t *)Malloc(sizeof(psema_t))) == NULL)
	  return NULL;

  theSemaphore->count = count;
  
  if((pthread_mutex_init(&theSemaphore->mutex, NULL)) != 0)
	  return NULL;
 
  if((pthread_cond_init(&theSemaphore->cond, NULL)) != 0)
	  return NULL;

  return (TNlmSemaphore)theSemaphore;


#elif defined(WIN32)
  HANDLE theSemaphore;

  if((theSemaphore = 
	  CreateSemaphore(NULL, count, MAX_SEMAPHORE_COUNT, NULL)) != NULL)
	  return (TNlmSemaphore)theSemaphore;
  else
	  return NULL;
#else
  return NULL;
#endif
}


/* ---------------------  NlmSemaDestroy  ---------------------------
   Purpose:     To destroy specified semaphore.
  
   Parameters:  theSemaphore - handle of the semaphore
  
   Returns:     Zero is returned when succesful. A non-zero value 
                indicates an error
  
   Description: destroys any state associated with the semaphore pointed
                to by theSemaphore
                
   NOTE:        The space for storing the semaphore is NOT freed.
   -----------------------------------------------------------------*/
Int4 NlmSemaDestroy(
		   TNlmSemaphore theSemaphore
    )
{
#if defined(SOLARIS_THREADS_AVAIL)
	
  Int4 error;
  
  error = sema_destroy((sema_t *)theSemaphore);  
  free(theSemaphore);
  return error;
  
#elif defined(POSIX_THREADS_AVAIL)
  
  psema_t *PSemaphore = (psema_t *)theSemaphore;
  
  if(pthread_mutex_destroy(&PSemaphore->mutex) != 0) {
    free(theSemaphore);
    return -1;
  }
    
  if(pthread_cond_destroy(&PSemaphore->cond) != 0) {
    free(theSemaphore);
    return -1;
  }
      
  free(theSemaphore);
  return 0;

#elif defined(WIN32)
  return CloseHandle((HANDLE)theSemaphore);
#else
  return 0;
#endif
}


/* ---------------------  NlmSemaWait  ------------------------------
   Purpose:      To wait until semaphore is not zero.
  
   Parameters:   theSemaphore - handle of the semaphore
  
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
  
   Description:  Blocks the calling thread until the count in the 
                 semaphore pointed by theSemaphore becomes greater than zero
                 and then atomically decrements it  
                
   NOTE:
   -----------------------------------------------------------------*/
Int4 NlmSemaWait(
		TNlmSemaphore theSemaphore
    )
{
#if defined(SOLARIS_THREADS_AVAIL)

  return sema_wait((sema_t *)theSemaphore);  

#elif defined(POSIX_THREADS_AVAIL)
 
  psema_t *PSemaphore = (psema_t *)theSemaphore;

  if((pthread_mutex_lock(&PSemaphore->mutex)) != 0)
	 return -1;

  if (PSemaphore->count > 0)
	  PSemaphore->count--;
  else {
	  do {
		  pthread_cond_wait(&PSemaphore->cond, &PSemaphore->mutex);
	  } while (PSemaphore->count == 0);
	  
	  PSemaphore->count--;
  }

  if((pthread_mutex_unlock(&PSemaphore->mutex)) != 0)
	  return -1;
  return 0;

#elif defined(WIN32)
  DWORD error;
  if((error = WaitForSingleObject((HANDLE)theSemaphore,INFINITE)) != WAIT_OBJECT_0)
	  return error;
  else
	  return 0;
  
#else
  return 0;
#endif
}


/* ---------------------  NlmSemaTryWait  ---------------------------
   Purpose:      To poll semaphore status    
  
   Parameters:   theSemaphore - handle of the semaphore to poll
  
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
  
   Description:  atomically decrements the count in the semaphore 
                 pointed to by theSemaphore if the count is greater
                 than zero. Otherwise it returns an error.
                
   NOTE:
  -----------------------------------------------------------------*/
Int4 NlmSemaTryWait(
		   TNlmSemaphore theSemaphore
    )
{
#if defined(SOLARIS_THREADS_AVAIL)

  return sema_trywait((sema_t *)theSemaphore);  

#elif defined (POSIX_THREADS_AVAIL)

  psema_t *PSemaphore = (psema_t *)theSemaphore;

  if((pthread_mutex_lock(&PSemaphore->mutex)) != 0)
    return -1;

  if (PSemaphore->count > 0) {
    PSemaphore->count--;
    if((pthread_mutex_unlock(&PSemaphore->mutex)) != 0)
      return -1;
    return 0;
  } else {
    if((pthread_mutex_unlock(&PSemaphore->mutex)) != 0)
      return -1;
    return -1;
  }

#elif defined(WIN32)

  DWORD error;
  if((error = WaitForSingleObject((HANDLE)theSemaphore, 0)) != WAIT_OBJECT_0)
	  return error;
  else
	  return 0;

#else
  return 0;
#endif
}


/* ---------------------  NlmSemaPost  ------------------------------
   Purpose:    To increase the count of semaphore by one
  
   Parameters: theSemaphore - handle of the semaphore to post
  
   Returns:    Zero is returned when succesful. A non-zero value
               indicates an error
  
   Description:  atomically increments the count semaphore pointed to by
                 thSemaphore, one is unblocked.
                
   NOTE
   -----------------------------------------------------------------*/
Int4 NlmSemaPost(
		TNlmSemaphore theSemaphore
    )
{
#if defined(SOLARIS_THREADS_AVAIL)

  return sema_post((sema_t *)theSemaphore);  


#elif defined (POSIX_THREADS_AVAIL)

  psema_t *PSemaphore = (psema_t *)theSemaphore;

  if((pthread_mutex_lock(&PSemaphore->mutex)) != 0)
	 return -1;

  PSemaphore->count++;

  pthread_cond_signal(&PSemaphore->cond);

  if((pthread_mutex_unlock(&PSemaphore->mutex)) != 0)
	 return -1;

  return 0;
  
#elif defined(WIN32)
  return ReleaseSemaphore((HANDLE)theSemaphore, 1, NULL);
#else
  return 0;
#endif
}

/********************************************************************/
/*                                                                  */
/* === Semaphore-like object to keep read/write syncronization ==== */
/*                                                                  */
/********************************************************************/

/* ---------------------  NlmRWlock  --------------------------------
   Purpose:     Initialize readers/writer lock   
  
   Parameters:  None
  
   Returns:     Handle of the new RW - lock or NULL if error occour
  
   Description: Readers/writer locks must be initialized before use. This
                function initialize new lock.
                
   NOTE: 
   -----------------------------------------------------------------*/
TNlmRWlock NlmRWinit(void)
{
#if defined(SOLARIS_THREADS_AVAIL)

  int status;
  rwlock_t *theRWlock;

  if((theRWlock = malloc(sizeof(rwlock_t))) == NULL)
    return NULL;
  
  if ((status = rwlock_init(theRWlock, USYNC_THREAD, NULL)) != 0)
    return NULL;
  else 
    return (TNlmRWlock *)theRWlock;  
#elif defined(POSIX_THREADS_AVAIL)

  rwlock_t *theRWlock; 
 	  
  if ((theRWlock = (rwlock_t *)Malloc(sizeof(rwlock_t))) == NULL)
    return NULL;

  theRWlock->readers = 0;
  
  if((pthread_mutex_init(&theRWlock->mutex, NULL)) != 0)
	  return NULL;
 
  if((pthread_cond_init(&theRWlock->cond_r, NULL)) != 0)
	  return NULL;

  if((pthread_cond_init(&theRWlock->cond_w, NULL)) != 0)
	  return NULL;

  return (TNlmRWlock)theRWlock;


#elif defined(WIN32)
  return NULL;
#else
  return NULL;
#endif
}



/* ---------------------  NlmRWdestroy  -----------------------------
   Purpose:      To destroy RW lock
  
   Parameters:   theRWlock - handle of the RW lock to destroy
  
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
  
   Description:  Destroys any state associated with the RW lock
                 pointed to by theRWlock.
                
   NOTE:         The space for storing the RWlock is not freed.

  -----------------------------------------------------------------*/
Int4 NlmRWdestroy(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)
	
	Int4 error;	
  
	error = rwlock_destroy((rwlock_t *)theRWlock);  
	free(theRWlock);
	return(error);

#elif defined (POSIX_THREADS_AVAIL)

  rwlock_t *RWlock = (rwlock_t *)theRWlock;
  
  if(pthread_mutex_destroy(&RWlock->mutex) != 0) {
    free(theRWlock);
    return -1;
  }
    
  if(pthread_cond_destroy(&RWlock->cond_r) != 0) {
    free(theRWlock);
    return -1;
  }

  if(pthread_cond_destroy(&RWlock->cond_w) != 0) {
    free(theRWlock);
    return -1;
  }
      
  free(theRWlock);
  return 0;

#elif defined(WIN32)
  return 0;
#else
  return 0;
#endif
}



/* ---------------------  NlmRWrdlock  ------------------------------
   Purpose:      To acquire a read lock    
  
   Parameters:   theRWlock - handle of the RW lock
  
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
  
   Description:  Acquires a read lock on the readers/writer lock
                 pointed to by theRWlock. If the RW lock is already
                 locked for writing, the calling thread blocks until 
                 the write lock is released.
                
   NOTE          More than one thread may hold a read lock on a RW lock at any 
                 one time.

  -----------------------------------------------------------------*/
Int4 NlmRWrdlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)

  return rw_rdlock((rwlock_t *)theRWlock);  

#elif defined(POSIX_THREADS_AVAIL)

  rwlock_t *PRWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&PRWlock->mutex)) != 0)
	 return -1;

  if (PRWlock->readers >= 0)
	  PRWlock->readers++; 
  else {
	  do {
		  pthread_cond_wait(&PRWlock->cond_r, &PRWlock->mutex);
	  } while (PRWlock->readers != 0);
	  
	  PRWlock->readers++;
  }

  if((pthread_mutex_unlock(&PRWlock->mutex)) != 0)
	  return -1;
  return 0;


#elif defined(WIN32)
  return 0;
#else
  return 0;
#endif
}


        
/* ---------------------  NlmRWwrlock  ------------------------------
   Purpose:      To acquire a write lock 
  
   Parameters:   theRWlock - handle of the RW lock
  
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
  
   Description:  Acquires a write lock on the readers/writer lock
                 pointed to by theRWlock. If the RW lock is already
                 locked for reading or writing, the calling thread blocks 
                 until the lock is released.
                
   NOTE:         Only one thread may hold a write lock on a RW lock at any 
                 one time.

  -----------------------------------------------------------------*/
Int4 NlmRWwrlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)

  return rw_wrlock((rwlock_t *)theRWlock);  

#elif defined(POSIX_THREADS_AVAIL)
  rwlock_t *PRWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&PRWlock->mutex)) != 0)
	 return -1;

  if (PRWlock->readers == 0)
	  PRWlock->readers--; /* -1 is special value for writer */
  else {
	  do {
		  pthread_cond_wait(&PRWlock->cond_w, &PRWlock->mutex);
	  } while (PRWlock->readers != 0);
	  
	  PRWlock->readers--;
  }

  if((pthread_mutex_unlock(&PRWlock->mutex)) != 0)
	  return -1;
  return 0;

#elif defined(WIN32)
  return 0;
#else
  return 0;
#endif
}


        
/* ---------------------  NlmRWunlock  ------------------------------
   Purpose:      To unlock the RW lock    
  
   Parameters:   theRWlock - handle of the RW lock
  
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
  
   Description:  The RW lock must be locked and the calling thread must 
                 hold the lock either for reading or for writing. If
                 any other threads are waiting for the RW lock to become 
                 available, one of them is unblocked.
               
   NOTE:         If the calling thread does not hold the lock for either 
                 reading or writing no error status is returned and the 
                 behavior of the program is undefined.

   -----------------------------------------------------------------*/
Int4 NlmRWunlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)

  return rw_unlock((rwlock_t *)theRWlock);  

#elif defined (POSIX_THREADS_AVAIL)

  rwlock_t *RWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&RWlock->mutex)) != 0)
	 return -1;

  if(RWlock->readers == -1) { /* Hmm.. this is writer unlock */
    RWlock->readers++;
    pthread_cond_broadcast(&RWlock->cond_r);
    pthread_cond_signal(&RWlock->cond_w);
  } else {
    RWlock->readers--;        /* Hmm.. this is reader unlock */
    if(!RWlock->readers) 
      pthread_cond_signal(&RWlock->cond_w); /* if we clear access to writer*/
  }
  
  if((pthread_mutex_unlock(&RWlock->mutex)) != 0)
	 return -1;

  return 0;

#elif defined(WIN32)
  return 0;
#else
  return 0;
#endif
}



/* ---------------------  NlmRWtryrdlock  ---------------------------
   Purpose:      To attempt to acquire read lock
  
   Parameters:   theRWlock - handle of the RW lock
  
   Returns:      Zero is returned when succesful. A non-zero value
                 indicates an error
  
   Description:  Attempts to acquire a read lock on the RW lock
                 pointed to by theRWlock. If the RW lock is already
                 locked for writing, it returnes an error. Otherwise
                 the read lock is acquired
                
   NOTE 
   -----------------------------------------------------------------*/
Int4 NlmRWtryrdlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)

  return rw_tryrdlock((rwlock_t *)theRWlock);  

#elif defined(POSIX_THREADS_AVAIL)

  rwlock_t *PRWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&PRWlock->mutex)) != 0)
	 return -1;

  if (PRWlock->readers >= 0)
    PRWlock->readers++; 
  else 
    return -1;
  
  if((pthread_mutex_unlock(&PRWlock->mutex)) != 0)
    return -1;
  return 0;
  
#elif defined(WIN32)
  return 0;
#else
  return 0;
#endif
}



/* ---------------------  NlmRWtrywrlock  ---------------------------
   Purpose:       To attempt to acquire write lock
  
   Parameters:    theRWlock - handle of the RW lock
  
   Returns:       Zero is returned when succesful. A non-zero value
                  indicates an error
  
   Description:   Attempts to acquire a write lock on the RW lock
                  pointed to by theRWlock. If the RW lock is already
                  locked for reading or writing, it returnes an error. 
                  Otherwise the write lock is acquired
                
   NOTE
   -----------------------------------------------------------------*/
Int4 NlmRWtrywrlock(TNlmRWlock theRWlock)
{
#if defined(SOLARIS_THREADS_AVAIL)

  return rw_trywrlock((rwlock_t *)theRWlock);  

#elif defined(POSIX_THREADS_AVAIL)

  rwlock_t *PRWlock = (rwlock_t *)theRWlock;

  if((pthread_mutex_lock(&PRWlock->mutex)) != 0)
	 return -1;

  if (PRWlock->readers = 0)
    PRWlock->readers--; 
  else 
    return -1;
  
  if((pthread_mutex_unlock(&PRWlock->mutex)) != 0)
    return -1;
  return 0;

#elif defined(WIN32)
  return 0;
#else
  return 0;
#endif
}


/* ---------------------  NlmMutexInit  ------------------------------
   Purpose:     To create and initialize new mutex
  
   Parameters:  None
  
   Returns:     Mutex handle
  
   Description: To create new mutex.
                
   NOTE:        This mutex can be used to synchronize threads in this
                process only.
   -----------------------------------------------------------------*/
TNlmMutex NlmMutexInit( void )
{
#if defined(SOLARIS_THREADS_AVAIL)
  mutex_t *theMutex = (mutex_t *)Malloc( sizeof(mutex_t) );

  if(theMutex == NULL  ||
     mutex_init(theMutex, NULL, NULL) != 0)
    return NULL;

  return (TNlmMutex)theMutex;

#elif defined(POSIX_THREADS_AVAIL)
  pthread_mutex_t *theMutex = (pthread_mutex_t *)
    Malloc( sizeof(pthread_mutex_t) );

  if(theMutex == NULL  ||
     pthread_mutex_init(theMutex, NULL) != 0)
    return NULL;

  return (TNlmMutex)theMutex;

#elif defined(WIN32)

  HANDLE theMutex;
  theMutex = CreateMutex(NULL, FALSE, NULL);
  return (TNlmMutex)theMutex;

#else
  return NULL;
#endif
}


/* ---------------------  NlmMutexLock  ------------------------------
   Purpose:      To lock mutex.
  
   Parameters:   theMutex - handle of the mutex
  
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
  
   Description:  Blocks code access
                
   NOTE:
   -----------------------------------------------------------------*/
Int4 NlmMutexLock(
		TNlmMutex theMutex
    )
{
#if defined(SOLARIS_THREADS_AVAIL)
  return mutex_lock((mutex_t *)theMutex);  

#elif defined(POSIX_THREADS_AVAIL)
  return pthread_mutex_lock((pthread_mutex_t *)theMutex); 

#elif defined(WIN32)
  
  DWORD error;
  
  if((error = WaitForSingleObject((HANDLE)theMutex, INFINITE)) != WAIT_OBJECT_0)
	  return error;
  else
	  return 0;

#else
  return 0;
#endif
}


/* ---------------------  NlmMutexUnlock  ------------------------------
   Purpose:      To unlock mutex.
  
   Parameters:   theMutex - handle of the mutex
  
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
  
   Description:  Unblocks code access
                
   NOTE:
   -----------------------------------------------------------------*/
Int4 NlmMutexUnlock(
		TNlmMutex theMutex
    )
{
#if defined(SOLARIS_THREADS_AVAIL)
  return mutex_unlock((mutex_t *)theMutex);  

#elif defined(POSIX_THREADS_AVAIL)
  return pthread_mutex_unlock((pthread_mutex_t *)theMutex);  

#elif defined(WIN32)
  return ReleaseMutex((HANDLE)theMutex);
#else
  return 0;
#endif
}


/* ---------------------  NlmMutexTryLock  ------------------------------
   Purpose:      To poll mutex.
  
   Parameters:   theMutex - handle of the mutex
  
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
  
   Description:  Poll condition of mutex - do not block if mutex busy
                
   NOTE:
   -----------------------------------------------------------------*/
Int4 NlmMutexTryLock(
		TNlmMutex theMutex
    )
{
#if   defined(SOLARIS_THREADS_AVAIL)
  return mutex_trylock((mutex_t *)theMutex);  

#elif defined(POSIX_THREADS_AVAIL)
  return pthread_mutex_trylock((pthread_mutex_t *)theMutex);  

#elif defined(WIN32)

  DWORD error;
  
  if((error = WaitForSingleObject((HANDLE)theMutex, 0)) != WAIT_OBJECT_0)
	  return error;
  else
	  return 0;

#else
  return 0;
#endif
}


/* ---------------------  NlmMutexDestroy ------------------------------
   Purpose:      To destroy mutex.
  
   Parameters:   theMutex - handle of the mutex
  
   Returns:      Zero is returned when successful. A non-zero value
                 indicates an error
  
   Description:  Remove mutex from execution.
                
   NOTE:
   -----------------------------------------------------------------*/
Int4 NlmMutexDestroy(
		TNlmMutex theMutex
    )
{
#if defined(SOLARIS_THREADS_AVAIL)
  Int4 status = mutex_destroy( (mutex_t *)theMutex );
  Free( theMutex );
  return status;

#elif defined(POSIX_THREADS_AVAIL)
  Int4 status = pthread_mutex_destroy( (pthread_mutex_t *)theMutex );
  Free( theMutex );
  return status;

#elif defined(WIN32)
  return CloseHandle((HANDLE)theMutex);
#else
  return 0;
#endif
}

/* ---------------------  NlmCPUNumber ------------------------------
   Purpose:      To get number of CPU on machine
  
   Parameters:   none
  
   Returns:      Number of CPU
  
   Description:  
   -----------------------------------------------------------------*/

Int4 NlmCPUNumber(void)

{
#if defined(WIN32)
  SYSTEM_INFO sysInfo;
  GetSystemInfo(&sysInfo);
  return sysInfo.dwNumberOfProcessors;
#elif defined(OS_UNIX_SYSV) && defined(SYSPROC_SPARC)  /* SPARC Solaris */
  return  sysconf(_SC_NPROCESSORS_ONLN);
#elif defined(OS_UNIX_SYSV) && defined(PROC_I80X86)  /* Intel Solaris */
  return  sysconf(_SC_NPROCESSORS_ONLN);
#elif  defined(PROC_MIPS) && defined(_SC_NPROC_ONLN)
  return sysconf(_SC_NPROC_ONLN);
#else
  return -1;
#endif
}
/* ---------------------  NlmThreadsAvailable ------------------------------
   Purpose:      Check to see if threads are available on this platform
  
   Parameters:   none
  
   Returns:      TRUE if threads are available, FALSE otherwise
  
   Description:  Check to see if threads are available on this platform
   -----------------------------------------------------------------*/
Boolean NlmThreadsAvailable(void)
{
#if   defined(SOLARIS_THREADS_AVAIL)
  return TRUE;
#elif defined(POSIX_THREADS_AVAIL)
  return TRUE;
#elif defined(WIN32)
  return TRUE;
#else
  return FALSE;
#endif
}


#ifdef TEST_MODULE_NCBITHR

/***********************************************************************
 *  TEST
 ***********************************************************************/

#if !defined(SOLARIS_THREADS_AVAIL)  &&  !defined(POSIX_THREADS_AVAIL)  &&  !defined(WIN32)

!!!!! ATTENTION -- BAD PLATFORM TO TEST THREADS  !!!!!

#endif


#include <stdio.h>

static void TEST__ThreadExit0(VoidPtr user_arg)
{
  fprintf(stderr, "ThreadExit0():  Self=%ld, Arg = %p\n",
          (long)NlmThreadSelf(), user_arg);
  Nlm_MemFree( user_arg );
}

static void TEST__ThreadExit1(VoidPtr user_arg)
{
  fprintf(stderr, "ThreadExit1():  Self=%ld, Arg = %ld\n",
          (long)NlmThreadSelf(), (long)user_arg);
}

static void TEST__ThreadExit2(VoidPtr user_arg)
{
  fprintf(stderr, "ThreadExit2():  Self=%ld, Arg = %ld\n",
          (long)NlmThreadSelf(), (long)user_arg);
}


TNlmThreadOnExit exit_func_arr[] =
{
  TEST__ThreadExit0,
  TEST__ThreadExit1,
  TEST__ThreadExit2,
  TEST__ThreadExit1,
  TEST__ThreadExit2,
  NULL
};

typedef struct
{
  Int4                  n_exit_func;
  TNlmThreadOnExit PNTR exit_func;
}  MyExitSet, PNTR MyExitSetPtr;


static VoidPtr TEST__MyThread(VoidPtr arg)
{
  MyExitSetPtr exit_func_arr = (MyExitSetPtr) arg;
  Int4 f;

  fprintf(stderr, "TEST__MyThread():  Thread #%ld started\n",
          (long)NlmThreadSelf());

  /* register the dummy(debug-print only) exit functions */
  for (f = exit_func_arr->n_exit_func - 1;  f > 0;  f--)
    {
      NlmThreadAddOnExit(exit_func_arr->exit_func[f],
                         (VoidPtr)NlmThreadSelf());

      if (f%2 == 0)
        NlmThreadRemoveOnExit(exit_func_arr->exit_func[f],
                              (VoidPtr)NlmThreadSelf());
    }

  /* register the exit function to deallocate dynamic memory */
  NlmThreadAddOnExit(exit_func_arr->exit_func[0], (VoidPtr)exit_func_arr);

  /* sleep for a while before exiting */
#if defined (OS_UNIX)
  sleep( 2 );
#elif defined (WIN32)
  Sleep( 2000 );
#endif

  return NULL;
}


static Int2 TEST__threads(Int4 n_threads)
{
  Int4 t;
  TNlmThread *threads = Nlm_MemNew(n_threads * sizeof(TNlmThread));

  for (t = 0;  t < n_threads;  t++)
    {
      MyExitSetPtr exit_set = (MyExitSetPtr) Nlm_MemNew( sizeof(MyExitSet) );
      exit_set->n_exit_func = t + 1;
      exit_set->exit_func   = exit_func_arr;

      threads[t] = NlmThreadCreate(&TEST__MyThread, (VoidPtr)exit_set, THR_RUN);
      fprintf(stdout, "TEST__threads():  Starting thread #%ld\n",
              (long)threads[t]);
    }

  for (t = 0;  t < n_threads;  t++)
    {
      VoidPtr status;
      Int4 err_code = NlmThreadJoin(threads[t], &status);
      if (err_code != 0)
        fprintf(stdout, "TEST__threads():  Cannot join thread #%ld;  \
error code = %ld\n",
                (long)threads[t], (long)err_code);

      fprintf(stdout, "TEST__threads():  Thread #%ld joined;  \
terminated, exit status = %p\n",
              (long)threads[t], status);
    }

  Nlm_MemFree( threads );  

  fprintf(stdout, "TEST__threads():  FINISHED\n");
  return 0;
}


Int2 main( void )
{
  Int2 exit_code = TEST__threads( 4 );
#if defined (WIN32)
  Sleep( 20000L );
#endif

  return exit_code;
}

#endif  /* TEST_MODULE_NCBITHR */


/* EOF */
